"""
molar http 客户端
"""

from enum import Enum
import math
from pathlib import Path
import random

from joblib import Parallel, delayed
import requests


from fyuneru.http_utils import find_labels, get_item_info, get_task_info
from fyuneru.lib import mkdir, read_json, tqdm_loguru, write_json


class MolarDomain(Enum):
    """
    域名
    """

    CN = "https://app.molardata.com"
    OTHER = "https://app.abaka.ai"


def init_export(
    export_config: dict,
    token: str,
    domain: str,
    n_jobs: int = 64,
    tmp_dir: Path = Path(".data/tmp"),
) -> dict:
    SESSION_POOL = [requests.Session() for _ in range(int(math.sqrt(n_jobs)))]
    task_id = export_config["taskId"]
    tmp_dir = tmp_dir / task_id
    mkdir(tmp_dir)
    task_info_tmp = tmp_dir / "task_info.json"
    if task_info_tmp.exists():
        task_config = read_json(task_info_tmp)
    else:
        while not (task_config := get_task_info(task_id, token, domain)):
            continue
        task_config = task_config["data"]
        write_json(data=task_config, file=task_info_tmp)
    item_id_s = export_config["exportMetadata"]["match"]["itemIds"]
    data_s = Parallel(n_jobs=n_jobs)(
        delayed(get_item)(
            task_id, item_id, token, domain, random.choice(SESSION_POOL), tmp_dir
        )
        for item_id in tqdm_loguru(item_id_s, desc="get item")
    )
    return {
        "task": task_config,
        "config": export_config,
        "data": data_s,
    }


def get_item(
    task_id: str,
    item_id: str,
    token: str,
    domain: str,
    session: requests.Session,
    tmp_dir: Path,
) -> dict:
    item_tmp = tmp_dir / f"{item_id}.json"
    if item_tmp.exists():
        return read_json(item_tmp)

    while not (
        item_info := get_item_info(
            item_id=item_id, token=token, domain=domain, session=session
        )
    ):
        continue
    while not (
        labels := find_labels(
            task_id=task_id,
            item_id=item_id,
            token=token,
            domain=domain,
            session=session,
        )
    ):
        continue
    info = item_info["data"]
    labels = labels["data"]
    item_dict = {
        "info": info,
        "labels": labels,
    }
    write_json(data=item_dict, file=item_tmp)
    return item_dict



"""平台导出工程"""

from dataclasses import dataclass, field
from pathlib import Path

from loguru import logger

from .lib import read_json
from .molar_client import MolarClient


@dataclass
class LabelConfig:
    """标签配置"""

    label: str
    alias: str
    drawType: str  # pylint: disable=invalid-name
    color: str
    key: str


@dataclass
class TaskConfig:
    """任务配置"""

    task_id: str
    export_id: str
    label_configs: dict[str, LabelConfig] = field(default_factory=dict)


@dataclass
class Frame:
    """帧"""

    index: int
    lens_index: int
    url: str
    urls: list[str] = field(default_factory=list)
    # x: width, y: height, z: depth
    size: list[float] = field(default_factory=list)


@dataclass
class Label:
    """标签"""

    label_id: int
    uid: str
    label: str
    draw_type: str
    frame_index: int
    points: list = field(default_factory=list)
    lens_index: int = field(default=0)
    group: int = field(default=0)
    attributes: dict = field(default_factory=dict)


@dataclass
class Item:
    """条目"""

    uid: str
    batch_uid: str
    task_config: TaskConfig
    frames: dict[int, Frame] = field(default_factory=dict)
    labels: dict[str, Label] = field(default_factory=dict)


def get_label_config(
    export_info: MolarClient.ExportInfo, task_info: MolarClient.TaskInfo
):
    """获取标签配置"""
    label_configs: dict[str, LabelConfig] = {}
    for label_config in task_info.label_config:
        label = label_config["label"]
        alias = label
        draw_type = label_config["drawType"]
        color = label_config["color"]
        key = label_config["key"]
        label_configs[alias] = LabelConfig(
            label=label, alias=alias, drawType=draw_type, color=color, key=key
        )
    if not export_info.task_alias:
        logger.info("Task alias not configured")
        return label_configs
    for label, values in export_info.task_alias.items():

        if not isinstance(values, dict) or "label" not in values:
            continue
        alias = values["label"]
        label_configs[label].alias = alias
        label_configs[alias] = label_configs.pop(label)
    return label_configs


def parse_iat_frames(
    item: MolarClient.ItemInfo,
) -> dict[int, Frame]:
    """解析条目"""
    frames: dict[int, Frame] = {}
    for idx, frame in enumerate(item.info["url"]):
        frame_url = frame
        size = [item.info["size"][idx]["width"], item.info["size"][idx]["height"]]
        len_index = 0
        frames[idx] = Frame(
            index=idx,
            lens_index=len_index,
            url=frame_url,
            size=size,
        )
    return frames


def parse_iat_labels(labels: list[MolarClient.LabelInfo]) -> dict[str, Label]:
    """解析标签"""
    export_labels: dict[str, Label] = {}
    for label in labels:
        label_id = label["data"]["id"]
        uid = label.uid
        label = label["data"]["label"]
        draw_type = label["data"]["drawType"]
        group = label["data"]["group"]
        attributes = label["data"].get("attributes", {})
        frame_index = label["data"]["frameIndex"]
        lens_index = label["data"]["lensIndex"]
        points = label["data"]["points"]
        export_labels[uid] = Label(
            label_id=label_id,
            uid=uid,
            label=label,
            draw_type=draw_type,
            frame_index=frame_index,
            lens_index=lens_index,
            group=group,
            attributes=attributes,
            points=points,
        )
    return export_labels


def get_items(task_config: TaskConfig, items: list[MolarClient.ItemInfo]):
    """获取条目"""
    export_items: list[Item] = []
    for item in items:
        item_uid = item.uid
        batch_uid = item.batch_id
        item_frames: dict[int, Frame] = parse_iat_frames(item=item)
        item_labels: dict[str, Label] = parse_iat_labels(labels=item.labels)
        export_items.append(
            Item(
                uid=item_uid,
                batch_uid=batch_uid,
                task_config=task_config,
                frames=item_frames,
                labels=item_labels,
            )
        )
    return export_items


def get_export_task_info(
    task_info: MolarClient.TaskInfo, export_info: MolarClient.ExportInfo
):
    """获取导出任务信息"""
    task_id = task_info.uid
    export_id = (
        export_info.origin_data.get("_id", None)
        or export_info.origin_data["exportTaskId"]
    )
    label_configs = get_label_config(export_info, task_info)
    return TaskConfig(
        task_id=task_id,
        export_id=export_id,
        label_configs=label_configs,
    )


def export(token: str, export_jsons_path: Path, domain: str) -> tuple[list[Item], int]:
    """导出数据"""
    molar_client = MolarClient(token=token, domain=domain)
    export_jsons = read_json(export_jsons_path)
    export_idx = len(export_jsons) - 1
    export_info, task_info, items = molar_client.get_export(
        export_config=export_jsons[export_idx], thread_num=64, dsn_cache=True
    )
    task_config = get_export_task_info(task_info=task_info, export_info=export_info)
    export_items = get_items(task_config=task_config, items=items)
    return export_items, export_idx



import pytest

from fyuneru.lib import hash


def test_hash():
    print("lib.hash")
    print(hash(b"", "blake3"))
    print(hash(b"123", "blake3"))
    print(hash(b"abc", "blake3"))
    print(len(hash(b"abc", "blake3")))
    print(len(hash(b"abcd", "blake3")))

    print(hash(b"abc", "md5"))
    # print(hash(b"abc", "sha1"))


test_hash()



"""
点云相关工具
"""

from copy import deepcopy
import re
from functools import reduce
from pathlib import Path
from typing import Generator, NamedTuple

import numpy as np
from returns.io import impure_safe
from returns.result import Failure, Result, Success, safe

from fyuneru.geometry3d import SElement, to_homogeneous_matrix


class PcdHeaderMeta(NamedTuple):
    data_start: int
    header_bytes: bytes


class PcdHeader(NamedTuple):
    """PCD 头结构"""

    VERSION: str
    FIELDS: list[str]
    SIZE: list[int]
    TYPE: list[str]
    COUNT: list[int]
    WIDTH: int
    HEIGHT: int
    VIEWPOINT: list[int]
    POINTS: int
    DATA: str


class PcdHeaderCol(NamedTuple):
    idx: int
    field: str
    type: str
    count: int
    size: int


class Pcd(NamedTuple):
    pc: np.ndarray
    header: PcdHeader


@impure_safe
def read_head_bytes(file: Path, num_bytes: int) -> bytes:
    """读取文件前num_bytes字节"""
    with file.open("rb") as f:
        return f.read(num_bytes)


@impure_safe
def get_pcd_head_meta(file: Path, start_bytes) -> PcdHeaderMeta:
    """
    取 PCD 头元信息

    支持的 DATA 格式：
        DATA ascii
        DATA binary
        DATA binary_compressed

    Args:
        file: PCD 文件路径
        cat_bytes: 读取头部的最大字节数

    Returns:
        PcdHeaderMeta
    """
    # 读取头部字节（解包 Result -> bytes）
    origin_head_bytes = read_head_bytes(file, start_bytes).unwrap()._inner_value

    # 正则匹配 DATA 行
    data_pattern = rb"DATA\s+(ascii|binary|binary_compressed)\s*\n"
    match = re.search(data_pattern, origin_head_bytes)
    if not match:
        raise ValueError("PCD ERROR: DATA line not found or invalid format")

    # 数据起始位置（换行后一个字节）
    data_start = match.end()

    return PcdHeaderMeta(
        data_start=data_start, header_bytes=origin_head_bytes[:data_start]
    )


def convert_pcd_header_to_str(pcd_header: PcdHeader) -> str:
    return (
        "# .PCD v0.7 - Point Cloud Data file format\n"
        f"VERSION {pcd_header.VERSION}\n"
        + f"FIELDS {' '.join(pcd_header.FIELDS)}\n"
        + f"SIZE {' '.join(map(str, pcd_header.SIZE))}\n"
        + f"TYPE {' '.join(pcd_header.TYPE)}\n"
        + f"COUNT {' '.join(map(str, pcd_header.COUNT))}\n"
        + f"WIDTH {pcd_header.WIDTH}\n"
        + f"HEIGHT {pcd_header.HEIGHT}\n"
        + f"VIEWPOINT {' '.join(map(str, pcd_header.VIEWPOINT))}\n"
        + f"POINTS {pcd_header.POINTS}\n"
        + f"DATA {pcd_header.DATA}\n"
    )


def convert_str_to_pcd_header(header_str: str) -> PcdHeader:
    HEADER_PATTERN = re.compile(
        r"^(VERSION|FIELDS|SIZE|TYPE|COUNT|WIDTH|HEIGHT|VIEWPOINT|POINTS|DATA)"
    )
    """将字符串转换为pcd头"""
    header_lines = [line.strip() for line in header_str.split("\n") if line.strip()]
    header = {}
    for line in header_lines[1:]:
        match = HEADER_PATTERN.match(line)
        if not match:
            raise ValueError(f"Invalid header line: {line}")
        parts = line.split()
        key = match.group(0)
        header[key] = parts[1:] if len(parts) > 2 else parts[1]
    header["POINTS"] = int(header["POINTS"])
    header["WIDTH"] = int(header["WIDTH"])
    header["HEIGHT"] = int(header["HEIGHT"])
    return PcdHeader(**header)


@safe
def parse_pcd_header(meta: PcdHeaderMeta) -> PcdHeader:
    """解析pcd头"""
    header_str = meta.header_bytes.decode(encoding="utf-8")
    return convert_str_to_pcd_header(header_str)


@safe
def build_dtype(
    fields: list[str], types: list[str], sizes: list[int], counts: list[int]
) -> np.dtype:
    """构建pcd numpy结构
    Args:
        fields (list): 字段名
        types (list): 类型
        sizes (list): 大小
        counts (list): 数量
    Raises:
        ValueError: 不支持的类型
    Returns:
        np.dtype: numpy结构
    """
    TYPE_MAP = {
        ("U", 1): np.uint8,
        ("U", 2): np.uint16,
        ("U", 4): np.uint32,
        ("U", 8): np.uint64,
        ("I", 1): np.int8,
        ("I", 2): np.int16,
        ("I", 4): np.int32,
        ("I", 8): np.int64,
        ("F", 4): np.float32,
        ("F", 8): np.float64,
    }
    field_dtypes = [TYPE_MAP[(t, int(s))] for t, s in zip(types, sizes)]
    dtype = reduce(
        lambda acc, zip_value: acc
        + [(f"{zip_value[0]}_{i}", zip_value[1]) for i in range(int(zip_value[2]))],
        zip(fields, field_dtypes, counts),
        [],
    )
    return np.dtype(dtype)


@impure_safe
def read_pcd(pcd_file: Path, start_bytes: int = 1024) -> Result[Pcd, ValueError]:
    """读取pcd文件
    Args:
        pcd_file: pcd文件路径
        start_bytes: 读取头部的最大字节数
    Returns:
        tuple[np.ndarray, PcdHeader]
    """
    meta = get_pcd_head_meta(pcd_file, start_bytes).unwrap()._inner_value
    header = parse_pcd_header(meta).unwrap()
    data_type = header.DATA.lower()
    data = None
    dtype = build_dtype(header.FIELDS, header.TYPE, header.SIZE, header.COUNT).unwrap()

    match data_type:
        case "binary":
            data = __read_pcd_binary(pcd_file, dtype, meta.data_start)
        case _:
            raise ValueError(f"Unsupported data type: {data_type}")

    return Pcd(pc=np.array(data.view(dtype).tolist()), header=header)


def __read_pcd_binary(file_path: Path, dtype: np.dtype, data_start: int) -> np.ndarray:
    return np.fromfile(file=file_path, dtype=dtype, offset=data_start)


def __write_pcd_binary(
    file: Path, header: PcdHeader
) -> Generator[None, np.ndarray, None]:
    """流式写入二进制PCD文件的生成器
    Args:
        file: PCD文件路径
        header: PCD头信息，包含字段定义

    Yields:
        None: 每次yield后等待接收点云数据
    """
    # 根据头部信息构建dtype
    dtype = build_dtype(header.FIELDS, header.TYPE, header.SIZE, header.COUNT).unwrap()

    with file.open("wb") as f:
        header_data = convert_pcd_header_to_str(header).encode("utf-8")
        f.write(header_data)

        while True:
            pc_data: np.ndarray = yield
            if pc_data is None:
                break
            f.write(array_to_structured(pc_data, dtype).tobytes(order="C"))


def array_to_structured(pc_array: np.ndarray, dtype: np.dtype) -> np.ndarray:
    """将二维数组转换为 build_dtype 构建的结构化数组"""
    N = pc_array.shape[0]
    arr = np.zeros(N, dtype=dtype)
    col_start = 0
    for name in dtype.names:
        field_dtype, _ = dtype.fields[name]  # 获取字段的 dtype 和偏移量
        # 判断这个字段是标量还是数组
        if field_dtype.shape == ():
            count = 1
        else:
            count = field_dtype.shape[0]
        cols = pc_array[:, col_start : col_start + count]
        if count == 1:
            # 标量字段（转成该字段基础类型）
            arr[name] = cols[:, 0].astype(field_dtype)
        else:
            # 向量字段（直接转成目标类型）
            arr[name] = cols.astype(field_dtype.base)
        col_start += count
    return arr


def get_pcd_writer(file: Path, header: PcdHeader):
    """获取已初始化的PCD写入器

    Args:
        file: PCD文件路径
        header: PCD头信息

    Returns:
        已初始化的写入器生成器，可直接使用send()方法写入数据
    """
    data_type = header.DATA.lower()
    match data_type:
        case "binary":
            writer = __write_pcd_binary(file, header)
            next(writer)  # 初始化写入器，写入头部信息
            return writer
        case _:
            raise ValueError(f"Unsupported data type: {data_type}")


def transform_pc(pc: np.ndarray, T: np.ndarray) -> np.ndarray:
    """
    三维点 转换到box坐标系
    """
    pc_homo = np.hstack([pc, np.ones((pc.shape[0], 1))])
    return (T @ pc_homo.T).T[:, :3]


def is_in_box_mask(
    pc: np.ndarray, box_center: SElement, box_size: list[float]
) -> np.ndarray:
    """
    返回布尔掩码，表示哪些点在box内。
    pc: (N, 3) ndarray
    box_center: 中心+旋转 (SElement)
    box_size: [xmin, xmax, ymin, ymax, zmin, zmax]（在box坐标系下）
    如果某个范围值为0，则不参与这方向的限制（相当于无限大）
    """
    # 世界坐标系 -> box局部坐标系
    T_glo_box = to_homogeneous_matrix(box_center)
    T_box_glo = np.linalg.inv(T_glo_box)
    pc_in_box_element = transform_pc(pc=pc, T=T_box_glo)

    # 处理 0 作为无限范围
    bounds = box_size.copy()
    if bounds[0] == 0:
        bounds[0] = -np.inf
    if bounds[1] == 0:
        bounds[1] = np.inf
    if bounds[2] == 0:
        bounds[2] = -np.inf
    if bounds[3] == 0:
        bounds[3] = np.inf
    if bounds[4] == 0:
        bounds[4] = -np.inf
    if bounds[5] == 0:
        bounds[5] = np.inf

    # 分别判断每个方向
    mask_x = (pc_in_box_element[:, 0] >= bounds[0]) & (
        pc_in_box_element[:, 0] <= bounds[1]
    )
    mask_y = (pc_in_box_element[:, 1] >= bounds[2]) & (
        pc_in_box_element[:, 1] <= bounds[3]
    )
    mask_z = (pc_in_box_element[:, 2] >= bounds[4]) & (
        pc_in_box_element[:, 2] <= bounds[5]
    )

    return mask_x & mask_y & mask_z


@safe
def add_column(pc: np.ndarray, col: np.ndarray) -> np.ndarray:
    return np.hstack([pc, col])


def append_pcd_header(origin_header: PcdHeader, col_header: PcdHeaderCol) -> PcdHeader:
    header = deepcopy(origin_header)
    header.FIELDS.append(col_header.field)
    header.COUNT.append(col_header.count)
    header.SIZE.append(col_header.size)
    header.TYPE.append(col_header.type)
    return header


def write_pcd(file: Path, header: PcdHeader, pc: np.ndarray):
    pcd_writer = get_pcd_writer(file=file, header=header)
    pcd_writer.send(pc)
    pcd_writer.close()



from enum import Enum
from functools import reduce
from typing import Callable


class NuScenesIndex(Enum):
    ATTRIBUTE = "attribute.json"
    # 标定token + 传感器 token信息
    CALIBRATED_SENSOR = "calibrated_sensor.json"
    CATEGORY = "category.json"
    EGO_POSE = "ego_pose.json"
    INSTANCE = "instance.json"
    LIDARSEG = "lidarseg.json"
    LOG = "log.json"
    MAP = "map.json"
    SAMPLE_ANNOTATION = "sample_annotation.json"
    # 多个sample data token + calibration token
    SAMPLE_DATA = "sample_data.json"
    # 时间戳+sample data token
    SAMPLE_TIME = "sample_time.json"
    SAMPLE = "sample.json"
    SCENE = "scene.json"
    # 传感器的token和channel名称
    SENSOR = "sensor.json"
    VISIBILITY = "visibility.json"
    SAMPLES_SUB = "samples"
    SWEEPS_SUB = "sweeps"


def sort_sample(sample_time: dict) -> list[str]:
    """
    排序sample索引
    """
    sorted_sample_time = sorted(sample_time.items(), key=lambda entry: int(entry[0]))
    return [sample_token for _, sample_token in sorted_sample_time]


def init_sample_data(sample_data_s: list[dict]) -> Callable[[str, str], dict]:
    """
    初始化sample信息
    """
    index_by_sample_sensor: dict[tuple[str, str], dict] = reduce(
        lambda acc, sample_data: acc.update(
            {
                (
                    sample_data["sample_token"],
                    sample_data["calibrated_sensor_token"],
                ): sample_data
            }
        )
        or acc,
        sample_data_s,
        dict(),
    )
    return lambda sample_token, calibrated_sensor_token: index_by_sample_sensor[
        (sample_token, calibrated_sensor_token)
    ]


def init_calibrated_sensor(calibrated_sensor_s: list[dict]) -> Callable[[str], dict]:
    """
    索引calibrated_sensor
    """
    index_by_calibrated_token: dict[str, dict] = reduce(
        lambda acc, calibrated_sensor: acc.update(
            {calibrated_sensor["token"]: calibrated_sensor}
        )
        or acc,
        calibrated_sensor_s,
        dict(),
    )
    return lambda calibrated_sensor_token: index_by_calibrated_token[
        calibrated_sensor_token
    ]


def init_sensor(sensor_s: list[dict]) -> Callable[[str], dict]:
    """
    索引sensor
    """
    index_by_sensor_token: dict[str, dict] = reduce(
        lambda acc, sensor: acc.update({sensor["channel"]: sensor}) or acc,
        sensor_s,
        dict(),
    )
    return lambda sensor_token: index_by_sensor_token[sensor_token]



from typing import NamedTuple

from fyuneru.lib import NonSerializable


class ImportItem(NamedTuple):
    info: any
    preData: list | NonSerializable


class CameraConfigV3(NamedTuple):
    name: str
    extrinsic: list
    intrinsic: list
    distortion: dict | NonSerializable
    projectionModel: str | NonSerializable


class D3V3Info(NamedTuple):
    urls: list[str]
    imgUrls: list | NonSerializable
    locations: list | NonSerializable
    cameraConfigs: list[CameraConfigV3] | NonSerializable


class Location(NamedTuple):
    name: str
    urls: list[str] | NonSerializable
    posMatrix: list[float]
